#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from typing import Any, Dict, List

from qgis.core import QgsGeometry

from quality_result_gui.api.types.quality_error import (
    QualityError,
    QualityErrorPriority,
    QualityErrorsByFeature,
    QualityErrorsByFeatureType,
    QualityErrorsByPriority,
    QualityErrorType,
)


@dataclass
class QualityErrorResponse:
    errors_by_priority: List[QualityErrorsByPriority] = field(init=False)
    _errors_obj: List[Dict[str, Any]]

    def __post_init__(self) -> None:
        self.errors_by_priority = self._parse_errors_by_priority(self._errors_obj)

    def _parse_errors_by_priority(
        self, errors_obj: List[Dict[str, Any]]
    ) -> List[QualityErrorsByPriority]:
        return [
            QualityErrorsByPriority(
                QualityErrorPriority(error_obj["priority"]),
                self._parse_errors_by_feature_type(
                    QualityErrorPriority(error_obj["priority"]), error_obj["errors"]
                ),
            )
            for error_obj in errors_obj
        ]

    def _parse_errors_by_feature_type(
        self, priority: QualityErrorPriority, errors_obj: List[Dict[str, Any]]
    ) -> List[QualityErrorsByFeatureType]:
        return [
            QualityErrorsByFeatureType(
                error_obj["feature_type"],
                self._parse_errors_by_feature(
                    priority, error_obj["feature_type"], error_obj["errors"]
                ),
            )
            for error_obj in errors_obj
        ]

    def _parse_errors_by_feature(
        self,
        priority: QualityErrorPriority,
        feature_type: str,
        errors_obj: List[Dict[str, Any]],
    ) -> List[QualityErrorsByFeature]:
        return [
            QualityErrorsByFeature(
                feature_type,
                error_obj["feature_id"],
                self._parse_errors(
                    priority, feature_type, error_obj["feature_id"], error_obj["errors"]
                ),
            )
            for error_obj in errors_obj
        ]

    def _parse_errors(
        self,
        priority: QualityErrorPriority,
        feature_type: str,
        feature_id: str,
        errors_obj: List[Dict[str, Any]],
    ) -> List[QualityError]:
        return [
            QualityError(
                priority,
                feature_type,
                feature_id,
                error_obj["error_id"],
                error_obj["unique_identifier"],
                QualityErrorType(error_obj["error_type"]),
                error_obj["attribute_name"],
                error_obj["error_description"],
                error_obj.get("extra_info", None),
                QgsGeometry.fromWkt(error_obj["wkt_geom"]),
                error_obj["is_user_processed"],
            )
            for error_obj in errors_obj
        ]



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Tuple, Union

from qgis.core import (
    QgsDrawSourceEffect,
    QgsEffectStack,
    QgsOuterGlowEffect,
    QgsProperty,
    QgsPropertyCollection,
    QgsSymbolLayer,
)
from qgis.PyQt.QtGui import QColor


def set_symbol_layer_data_defined_property_expressions(
    symbol_layer: QgsSymbolLayer, data_defined_property_expressions: Dict[str, str]
) -> None:
    """
    Sets the symbol layer data defined properties with expressions.

    Will silently ignore if the property name is not found on the symbol layer.
    """
    data_defined_properties = QgsPropertyCollection()

    property_name_to_number_mapping = {
        definition.name(): number
        for number, definition in symbol_layer.propertyDefinitions().items()
    }

    for property_name, property_expression in data_defined_property_expressions.items():
        property = QgsProperty.fromExpression(property_expression)
        if (
            property_number := property_name_to_number_mapping.get(property_name)
        ) is not None:
            data_defined_properties.setProperty(property_number, property)

    symbol_layer.setDataDefinedProperties(data_defined_properties)


def set_symbol_layer_simple_outer_glow_effect(
    symbol_layer: QgsSymbolLayer,
    color_rgba: str,
    spread: Tuple[float, str] = (2.0, "MM"),
    blur: Tuple[float, str] = (2.0, "MM"),
    opacity: float = 1,
) -> None:
    """
    Sets an outer glow effect on a symbol layer.
    """
    spread_amount, spread_unit = spread
    blur_amount, blur_unit = blur

    source_effect_layer = QgsDrawSourceEffect.create({"enabled": "1"})
    glow_effect_layer = QgsOuterGlowEffect.create(
        {
            "enabled": "1",
            "single_color": color_rgba,
            "spread": str(spread_amount),
            "spread_unit": spread_unit,
            "blur_level": str(blur_amount),
            "blur_unit": blur_unit,
            "opacity": str(opacity),
        }
    )

    paint_effect_stack = QgsEffectStack()
    paint_effect_stack.appendEffect(source_effect_layer)
    paint_effect_stack.appendEffect(glow_effect_layer)

    symbol_layer.setPaintEffect(paint_effect_stack)


def get_color(
    hex_or_rgb: Union[str, Tuple[int, int, int]], opacity: int = 100
) -> QColor:
    color = QColor(hex_or_rgb)
    color.setAlpha(int(opacity / 100 * 255))
    return color



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from qgis.core import QgsApplication
from qgis.gui import QgsGui
from qgis.PyQt import uic
from qgis.PyQt.QtCore import pyqtSignal
from qgis.PyQt.QtWidgets import QDockWidget, QShortcut
from qgis_plugin_tools.tools.i18n import tr

from quality_result_gui.ui.quality_errors_tree_filter_menu import (
    QualityErrorsTreeFilterMenu,
)

if TYPE_CHECKING:
    from qgis.PyQt.QtGui import QCloseEvent
    from qgis.PyQt.QtWidgets import QCheckBox, QLabel, QToolButton, QWidget

    from quality_result_gui.ui.quality_error_tree_view import QualityErrorTreeView


LOGGER = logging.getLogger(__name__)

DOCK_UI: "QWidget"
DOCK_UI, _ = uic.loadUiType(
    str(Path(__file__).parent.joinpath("quality_errors_dock.ui"))
)


class QualityErrorsDockWidget(QDockWidget, DOCK_UI):
    """
    Graphical user interface for quality errors dock widget.
    """

    SHORTCUT_TOGGLE_ERRORS_ON_MAP_FILTER = "Alt+Q"

    closed = pyqtSignal()

    # type necessary widgets that are provided from the .ui
    error_tree_view: "QualityErrorTreeView"
    filter_button: "QToolButton"
    info_label: "QLabel"
    filter_with_map_extent_check_box: "QCheckBox"
    show_errors_on_map_check_box: "QCheckBox"
    show_user_processed_errors_check_box: "QCheckBox"

    def __init__(self, parent: Optional["QWidget"] = None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self.shortcut_for_toggle_errors = QShortcut(self)
        self.shortcut_for_toggle_errors.setObjectName(
            "Toggle show errors on map filter"
        )
        self.shortcut_for_toggle_errors.setWhatsThis(
            tr("Toggle show errors on map filter")
        )
        self.shortcut_for_toggle_errors.activated.connect(
            self.show_errors_on_map_check_box.toggle
        )

        self.filter_menu = QualityErrorsTreeFilterMenu(self)
        self.filter_button.setIcon(QgsApplication.getThemeIcon("/mActionFilter2.svg"))
        self.filter_button.setMenu(self.filter_menu)

        self._update_filter_menu_icon_state()

    def _update_filter_menu_icon_state(self) -> None:
        if self.filter_menu.is_any_filter_active():
            self.filter_button.setDown(True)
        else:
            self.filter_button.setDown(False)

    def _register_shortcut(self) -> None:
        QgsGui.shortcutsManager().registerShortcut(
            self.shortcut_for_toggle_errors,
            self.SHORTCUT_TOGGLE_ERRORS_ON_MAP_FILTER,
        )

    def show(self) -> None:
        self._register_shortcut()
        return super().show()

    def closeEvent(self, event: "QCloseEvent") -> None:  # noqa: N802 (qt override)
        QgsGui.shortcutsManager().unregisterShortcut(self.shortcut_for_toggle_errors)

        self.closed.emit()
        return super().closeEvent(event)



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import logging
from pathlib import Path
from typing import Optional

from qgis.gui import QgsFileWidget
from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import QDialog, QDialogButtonBox, QPushButton, QWidget

FORM_CLASS: QWidget
FORM_CLASS, _ = uic.loadUiType(
    str(Path(__file__).parent.joinpath("dev_tools_dialog.ui"))
)


LOGGER = logging.getLogger(__name__)


class DevToolsDialog(QDialog, FORM_CLASS):

    quality_errors_data_file_widget: QgsFileWidget
    btn_open_quality_errors_dialog: QPushButton

    button_box: QDialogButtonBox

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        # Setup for quality errors dialog:
        self.quality_errors_data_file_widget.setDefaultRoot(
            str((Path(__file__).parent / "example_quality_errors").resolve())
        )
        self.quality_errors_data_file_widget.setFilter("*.json")
        self.quality_errors_data_file_widget.fileChanged.connect(
            self._enable_open_quality_errors_dialog_button
        )
        self.btn_open_quality_errors_dialog.setEnabled(False)
        self.btn_open_quality_errors_dialog.clicked.connect(self.accept)

    def _enable_open_quality_errors_dialog_button(self) -> None:
        self.btn_open_quality_errors_dialog.setEnabled(True)



#  Copyright (C) 2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from importlib.resources import as_file, files
from typing import TYPE_CHECKING, List, Optional, Union

from qgis.core import (
    QgsCentroidFillSymbolLayer,
    QgsFillSymbol,
    QgsGeometryGeneratorSymbolLayer,
    QgsLineSymbol,
    QgsMarkerSymbol,
    QgsSimpleFillSymbolLayer,
    QgsSimpleLineSymbolLayer,
    QgsSimpleMarkerSymbolLayer,
    QgsSvgMarkerSymbolLayer,
    QgsSymbolLayer,
)
from qgis.PyQt.QtGui import QColor

from quality_result_gui import resources
from quality_result_gui.api.types.quality_error import QualityErrorPriority
from quality_result_gui.style.quality_layer_error_symbol import ErrorSymbol
from quality_result_gui.utils import styling_utils
from quality_result_gui.utils.styling_utils import (
    set_symbol_layer_data_defined_property_expressions,
    set_symbol_layer_simple_outer_glow_effect,
)

if TYPE_CHECKING:
    from quality_result_gui.quality_error_visualizer import QualityError


# Colors for error features
FATAL_PRIMARY_COLOR = "#C31266"
FATAL_SECONDARY_COLOR = "#F24726"
WARNING_PRIMARY_COLOR = "#E65A13"
WARNING_SECONDARY_COLOR = "#FF8303"
INFO_PRIMARY_COLOR = "#EAA613"
INFO_SECONDARY_COLOR = "#FFFF00"

# Colors for selected error features
HIGHLIGHTED_FATAL_PRIMARY_COLOR = "#D90368"
HIGHLIGHTED_FATAL_SECONDARY_COLOR = "#F24726"
HIGHLIGHTED_WARNING_PRIMARY_COLOR = "#FB6107"
HIGHLIGHTED_WARNING_SECONDARY_COLOR = "#FF8303"
HIGHLIGHTED_INFO_PRIMARY_COLOR = "#FBD00E"
HIGHLIGHTED_INFO_SECONDARY_COLOR = "#FFFF00"


@dataclass
class QualityLayerColors:
    stroke_color: Union[QColor, str]
    secondary_color: Union[QColor, str]
    fill_color: Optional[Union[QColor, str]] = None


@dataclass
class QualityLayerStyle:
    colors_by_priority: dict[QualityErrorPriority, QualityLayerColors]
    line_width: float
    polygon_border_width: float
    marker_border_width: float
    marker_size: float


COLORS_FOR_ERRORS = {
    QualityErrorPriority.FATAL: QualityLayerColors(
        stroke_color=FATAL_PRIMARY_COLOR,
        secondary_color=styling_utils.get_color(FATAL_SECONDARY_COLOR, opacity=30),
    ),
    QualityErrorPriority.WARNING: QualityLayerColors(
        stroke_color=WARNING_PRIMARY_COLOR,
        secondary_color=styling_utils.get_color(WARNING_SECONDARY_COLOR, opacity=30),
    ),
    QualityErrorPriority.INFO: QualityLayerColors(
        stroke_color=INFO_PRIMARY_COLOR,
        secondary_color=styling_utils.get_color(INFO_SECONDARY_COLOR, opacity=30),
    ),
}


COLORS_FOR_HIGHLIGHTED_ERRORS = {
    QualityErrorPriority.FATAL: QualityLayerColors(
        stroke_color=HIGHLIGHTED_FATAL_PRIMARY_COLOR,
        fill_color=styling_utils.get_color(HIGHLIGHTED_FATAL_PRIMARY_COLOR, opacity=40),
        secondary_color=styling_utils.get_color(
            HIGHLIGHTED_FATAL_SECONDARY_COLOR, opacity=30
        ),
    ),
    QualityErrorPriority.WARNING: QualityLayerColors(
        stroke_color=HIGHLIGHTED_WARNING_PRIMARY_COLOR,
        fill_color=styling_utils.get_color(
            HIGHLIGHTED_WARNING_PRIMARY_COLOR, opacity=40
        ),
        secondary_color=styling_utils.get_color(
            HIGHLIGHTED_WARNING_SECONDARY_COLOR, opacity=30
        ),
    ),
    QualityErrorPriority.INFO: QualityLayerColors(
        stroke_color=HIGHLIGHTED_INFO_PRIMARY_COLOR,
        fill_color=styling_utils.get_color(HIGHLIGHTED_INFO_PRIMARY_COLOR, opacity=40),
        secondary_color=styling_utils.get_color(
            HIGHLIGHTED_INFO_SECONDARY_COLOR, opacity=30
        ),
    ),
}


class DefaultErrorSymbol(ErrorSymbol):
    SYMBOL_MAP_SCALE = 10000

    def __init__(
        self,
        quality_error: "QualityError",
    ) -> None:

        self.style: QualityLayerStyle = QualityLayerStyle(
            COLORS_FOR_ERRORS,
            line_width=1.2,
            polygon_border_width=0.4,
            marker_border_width=0.4,
            marker_size=5,
        )
        self.style_for_highlighted_error = QualityLayerStyle(
            COLORS_FOR_HIGHLIGHTED_ERRORS,
            line_width=1.6,
            polygon_border_width=0.8,
            marker_border_width=0.4,
            marker_size=5,
        )

        self.current_style = self.style

        self.priority = quality_error.priority

        self.icon_symbol_enabled_expression = f"@map_scale > {self.SYMBOL_MAP_SCALE}"
        self.geometry_symbol_enabled_expression = (
            f"@map_scale <= {self.SYMBOL_MAP_SCALE}"
        )

    def _get_polygon_symbol(self, highlighted: bool) -> QgsFillSymbol:
        style = self.style
        if highlighted:
            style = self.style_for_highlighted_error

        colors = style.colors_by_priority[self.priority]

        fill_color = QColor(0, 0, 0, 0)
        if colors.fill_color is not None:
            fill_color = QColor(colors.fill_color)

        symbol = QgsFillSymbol()
        fill_symbol_layer = QgsSimpleFillSymbolLayer.create(
            {
                "color": self.color_to_rgba_string(fill_color),
                "outline_style": "no",
            }
        )
        symbol.changeSymbolLayer(0, fill_symbol_layer)

        primary_border_symbol_layer = QgsSimpleLineSymbolLayer.create(
            {
                "line_color": self.color_to_rgba_string(colors.stroke_color),
                "line_width": str(style.polygon_border_width),
                "line_width_unit": "MM",
            }
        )
        symbol.appendSymbolLayer(primary_border_symbol_layer)

        secondary_border_symbol_layer = QgsSimpleLineSymbolLayer.create(
            {
                "line_color": self.color_to_rgba_string(colors.secondary_color),
                "line_width": "3.4",
                "line_width_unit": "MM",
                "joinstyle": "round",
            }
        )
        symbol.appendSymbolLayer(secondary_border_symbol_layer)

        priority_symbol_layer = self._create_priority_symbol_layer(self.priority)
        marker_symbol = QgsMarkerSymbol()
        marker_symbol.changeSymbolLayer(0, priority_symbol_layer)

        centroid_layer = QgsCentroidFillSymbolLayer.create({})
        centroid_layer.setSubSymbol(marker_symbol)

        symbol.appendSymbolLayer(centroid_layer)

        self._set_enabled_expression(
            centroid_layer,
            [
                fill_symbol_layer,
                primary_border_symbol_layer,
                secondary_border_symbol_layer,
            ],
        )

        return symbol

    def _get_line_symbol(self, highlighted: bool) -> QgsLineSymbol:
        style = self.style
        if highlighted:
            style = self.style_for_highlighted_error

        colors = style.colors_by_priority[self.priority]
        modified_stroke_color = QColor(colors.stroke_color)
        modified_stroke_color.setAlpha(170)

        line_symbol = QgsLineSymbol()

        primary_border_symbol_layer = QgsSimpleLineSymbolLayer.create(
            {
                "line_color": self.color_to_rgba_string(modified_stroke_color),
                "line_width": str(style.line_width),
                "line_width_unit": "MM",
            }
        )
        line_symbol.changeSymbolLayer(0, primary_border_symbol_layer)

        secondary_border_symbol_layer = QgsSimpleLineSymbolLayer.create(
            {
                "line_color": self.color_to_rgba_string(colors.secondary_color),
                "line_width": "3.4",
                "line_width_unit": "MM",
                "joinstyle": "round",
            }
        )
        line_symbol.appendSymbolLayer(secondary_border_symbol_layer)

        priority_symbol_layer = self._create_priority_symbol_layer(self.priority)
        marker_symbol = QgsMarkerSymbol()
        marker_symbol.changeSymbolLayer(0, priority_symbol_layer)

        generator_layer = QgsGeometryGeneratorSymbolLayer.create({})
        generator_layer.setGeometryExpression(
            "line_interpolate_point( $geometry, length3D($geometry) / 2)"
        )
        generator_layer.setSubSymbol(marker_symbol)

        if not line_symbol.appendSymbolLayer(generator_layer):
            raise ValueError("invalid line symbol priority_symbol_layer")
        self._set_enabled_expression(
            generator_layer,
            [secondary_border_symbol_layer, primary_border_symbol_layer],
        )

        return line_symbol

    def _get_point_symbol(self, highlighted: bool) -> QgsMarkerSymbol:
        style = self.style
        if highlighted:
            style = self.style_for_highlighted_error

        colors = style.colors_by_priority[self.priority]

        fill_symbol_layer = QgsSimpleMarkerSymbolLayer.create(
            {
                "size": str(style.marker_size),
                "size_unit": "MM",
                "color": self.color_to_rgba_string(QColor(0, 0, 0, 0)),
                "line_color": self.color_to_rgba_string(colors.stroke_color),
                "line_width": str(style.marker_border_width),
                "line_width_unit": "MM",
            }
        )
        modified_color = QColor(colors.secondary_color)
        modified_color.setAlphaF(0.7)
        set_symbol_layer_simple_outer_glow_effect(
            fill_symbol_layer, self.color_to_rgba_string(modified_color)
        )

        symbol = QgsMarkerSymbol()
        symbol.changeSymbolLayer(0, fill_symbol_layer)

        priority_symbol_layer = self._create_priority_symbol_layer(self.priority)
        symbol.appendSymbolLayer(priority_symbol_layer)
        self._set_enabled_expression(priority_symbol_layer, [fill_symbol_layer])

        return symbol

    def _set_enabled_expression(
        self,
        priority_symbol_layer: QgsSymbolLayer,
        geometry_layers: List[QgsSymbolLayer],
    ) -> None:
        if (
            self.icon_symbol_enabled_expression is not None
            and self.geometry_symbol_enabled_expression is not None
        ):
            set_symbol_layer_data_defined_property_expressions(
                priority_symbol_layer,
                {"enabled": self.icon_symbol_enabled_expression},
            )
            for layer in geometry_layers:
                set_symbol_layer_data_defined_property_expressions(
                    layer,
                    {"enabled": self.geometry_symbol_enabled_expression},
                )

    def _create_priority_symbol_layer(
        self, priority: QualityErrorPriority
    ) -> QgsSvgMarkerSymbolLayer:

        file_path = files(resources).joinpath("icons")

        if priority == QualityErrorPriority.FATAL:
            file_path = file_path.joinpath("quality_error_fatal.svg")
        elif priority == QualityErrorPriority.WARNING:
            file_path = file_path.joinpath("quality_error_warning.svg")
        elif priority == QualityErrorPriority.INFO:
            file_path = file_path.joinpath("quality_error_info.svg")
        else:
            raise ValueError(f"Unknown priority {str(priority)}")

        style = dict({})
        with as_file(file_path) as svg_file:
            style["name"] = str(svg_file)

        return QgsSvgMarkerSymbolLayer.create(style)



#  Copyright (C) 2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.


from typing import TYPE_CHECKING, Optional, cast

from qgis.gui import QgisInterface
from qgis.PyQt.QtCore import QObject, pyqtSignal
from qgis.utils import iface as utils_iface
from qgis_plugin_tools.tools.i18n import tr

from quality_result_gui import SelectionType
from quality_result_gui.api.types.quality_error import QualityError
from quality_result_gui.quality_data_fetcher import (
    CHECK_STATUS_LABELS,
    BackgroundQualityResultsFetcher,
    CheckStatus,
)
from quality_result_gui.quality_error_visualizer import QualityErrorVisualizer
from quality_result_gui.quality_errors_filters import (
    AbstractQualityErrorFilter,
    AttributeFilter,
    ErrorTypeFilter,
    FeatureTypeFilter,
)
from quality_result_gui.quality_errors_tree_model import (
    FilterByExtentProxyModel,
    FilterByShowUserProcessedProxyModel,
    FilterProxyModel,
    QualityErrorsTreeBaseModel,
    StyleProxyModel,
)
from quality_result_gui.ui.quality_errors_dock import QualityErrorsDockWidget

if TYPE_CHECKING:
    from quality_result_gui.api.quality_api_client import QualityResultClient
    from quality_result_gui.configuration import QualityLayerStyleConfig

iface = cast(QgisInterface, utils_iface)


class QualityResultManager(QObject):
    closed = pyqtSignal()
    error_checked = pyqtSignal(str, bool)
    quality_error_selected = pyqtSignal(QualityError, SelectionType)

    def __init__(
        self,
        api_client: "QualityResultClient",
        parent: Optional[QObject] = None,
        style_config: Optional["QualityLayerStyleConfig"] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._api_client = api_client

        self.visualizer = QualityErrorVisualizer(
            self._api_client.get_crs(), style_config
        )
        self.visualizer.show_errors()

        self.dock_widget = QualityErrorsDockWidget(iface.mainWindow())
        self.dock_widget.closed.connect(self.closed)

        self.dock_widget.error_tree_view.errors_inserted.connect(
            lambda errors: self.visualizer.add_new_errors(errors)
        )
        self.dock_widget.error_tree_view.errors_removed.connect(
            lambda errors: self.visualizer.remove_errors(errors)
        )

        self._fetcher = BackgroundQualityResultsFetcher(self._api_client, self)
        self._fetcher.status_changed.connect(self._update_info_label)

        self._base_model = QualityErrorsTreeBaseModel()
        self._base_model.error_checked.connect(self.error_checked)
        self._fetcher.results_received.connect(self._base_model.refresh_model)

        self._filter_model = FilterProxyModel()
        self._filter_model.setSourceModel(self._base_model)

        self._base_model.filterable_data_changed.connect(
            self._filter_model.invalidateFilter
        )

        self._filter_user_processed_model = FilterByShowUserProcessedProxyModel()
        self._filter_user_processed_model.setSourceModel(self._filter_model)

        self._filter_map_extent_model = FilterByExtentProxyModel()
        self._filter_map_extent_model.setSourceModel(self._filter_user_processed_model)

        self._styled_model = StyleProxyModel()
        self._styled_model.setSourceModel(self._filter_map_extent_model)

        self.dock_widget.error_tree_view.setModel(self._styled_model)

        self.dock_widget.show_user_processed_errors_check_box.toggled.connect(
            self._filter_user_processed_model.set_show_processed_errors
        )
        self.dock_widget.filter_with_map_extent_check_box.toggled.connect(
            self._filter_map_extent_model.set_enabled
        )
        self.dock_widget.show_errors_on_map_check_box.toggled.connect(
            self.visualizer.toggle_visibility
        )

        self.dock_widget.error_tree_view.quality_error_selected.connect(
            self.quality_error_selected
        )
        self.dock_widget.error_tree_view.quality_error_selected.connect(
            self.visualizer.on_error_selected
        )

        self._add_predefined_filters()

    def _add_predefined_filters(self) -> None:
        self._error_type_filter = ErrorTypeFilter()
        self.add_filter(self._error_type_filter)

        self._feature_type_filter = FeatureTypeFilter()
        self.add_filter(self._feature_type_filter)
        self._fetcher.results_received.connect(
            self._feature_type_filter.update_filter_from_errors
        )

        self._attribute_filter = AttributeFilter()
        self.add_filter(self._attribute_filter)
        self._fetcher.results_received.connect(
            self._attribute_filter.update_filter_from_errors
        )

    def unload(self) -> None:
        self._fetcher.stop()
        self._filter_map_extent_model.set_enabled(False)
        self.dock_widget.close()
        self.dock_widget.deleteLater()
        self.visualizer.remove_quality_error_layer()

    def _update_info_label(self, status: CheckStatus) -> None:
        status_text = CHECK_STATUS_LABELS.get(
            status, tr("Status of fetching quality result unknown")
        )
        self.dock_widget.info_label.setText(status_text)

    def show_dock_widget(self) -> None:
        self._fetcher.start()
        self.dock_widget.show()
        self.visualizer.initialize_quality_error_layer(
            self.dock_widget.show_errors_on_map_check_box.isChecked()
        )
        self.visualizer.add_new_errors(
            self.dock_widget.error_tree_view.get_all_quality_errors()
        )

    def hide_dock_widget(self) -> None:
        self._fetcher.stop()
        self.dock_widget.hide()
        self.visualizer.remove_quality_error_layer()

    def add_filter(self, filter: AbstractQualityErrorFilter) -> None:
        filter.filters_changed.connect(self.dock_widget._update_filter_menu_icon_state)
        self.dock_widget.filter_menu.add_filter_menu(filter.menu)
        self._filter_model.add_filter(filter)



#  Copyright (C) 2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.


import bisect
from abc import abstractmethod
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Hashable, List, Optional, Set, cast

from qgis.PyQt.QtCore import QObject, QSignalBlocker, pyqtSignal
from qgis.PyQt.QtGui import QMouseEvent
from qgis.PyQt.QtWidgets import QAction, QMenu
from qgis_plugin_tools.tools.i18n import tr

from quality_result_gui.api.types.quality_error import ERROR_TYPE_LABEL, QualityError
from quality_result_gui.quality_errors_tree_model import QualityErrorTreeItemType

if TYPE_CHECKING:
    from qgis.PyQt.QtWidgets import QWidget

    from quality_result_gui.api.types.quality_error import QualityErrorsByPriority

FEATURE_TYPE_FILTER_MENU_LABEL = tr("Feature type")
ERROR_TYPE_FILTER_MENU_LABEL = tr("Error type")
ATTRIBUTE_NAME_FILTER_MENU_LABEL = tr("Attribute Filter")


class FilterMenu(QMenu):
    """A QMenu for checkable filter actions with support for select all section and
    sorting"""

    def __init__(self, title: str, parent: Optional["QWidget"] = None) -> None:
        super().__init__(title, parent=parent)

        self._select_all_section_enabled = False
        self._sorted = False

        self._filter_actions: List[QAction] = []

    def mouseReleaseEvent(self, e: QMouseEvent) -> None:  # noqa: N802 (qt override)
        if not self.activeAction() or not self.activeAction().isEnabled():
            super().mouseReleaseEvent(e)
        else:
            with QSignalBlocker(self.activeAction()):
                self.activeAction().setEnabled(False)
                super().mouseReleaseEvent(e)
                self.activeAction().setEnabled(True)
            self.activeAction().trigger()

    def set_sorted(self, sorted: bool) -> None:
        """Sets the menu actions in sorted or not sorted order

        Args:
            sorted (bool): If True then menu actions are and will be sorted.
              If False menu actions are not kept sorted anymore.
        """

        if sorted == self._sorted:
            return

        if sorted:
            self._filter_actions.sort(key=lambda action: action.text())
            for action in self._filter_actions:
                self.removeAction(action)
                self.addAction(action)
            self._sorted = True
        else:
            self._sorted = False

    def set_select_all_section_enabled(self, enabled: bool) -> None:
        """Add or removes the (de)select all section

        Args:
            enabled (bool): If True then the section is added if not present already.
              The section is removed on False.
        """

        if enabled == self._select_all_section_enabled:
            return

        if enabled:
            self._add_select_all_section()
        else:
            self._remove_select_all_section()

    def _add_select_all_section(self) -> None:
        """Add a section to the menu with Select and Deselect all actions

        The section is added always on top of the menu
        """

        if self._select_all_section_enabled:
            return

        select_all_action = QAction(tr("Select all"), self)
        select_all_action.triggered.connect(self._select_all)

        deselect_all_action = QAction(tr("Deselect all"), self)
        deselect_all_action.triggered.connect(self._deselect_all)

        existing_actions = self.actions()
        first_action = existing_actions[0] if existing_actions else None
        separator = self.insertSeparator(first_action)  # type: ignore
        self.insertAction(separator, deselect_all_action)
        self.insertAction(deselect_all_action, select_all_action)

        self._select_all_section_enabled = True

    def _remove_select_all_section(self) -> None:
        """Removes the Select and Deselect all section"""

        if not self._select_all_section_enabled:
            return

        # (de)select all actions are always the first three
        for action in self.actions()[:3]:
            action.deleteLater()
            self.removeAction(action)

        self._select_all_section_enabled = False

    def _select_all(self) -> None:
        """Select all the user added checkable actions"""

        for action in self._filter_actions:
            action.setChecked(True)

    def _deselect_all(self) -> None:
        """Deselects all the user added checkable actions"""

        for action in self._filter_actions:
            action.setChecked(False)

    def remove_user_actions(self) -> None:
        """Removes all the user added actions"""
        while self._filter_actions:
            self.remove_filter_action(self._filter_actions[0])

    def add_checkable_action(self, label: str) -> QAction:
        """Adds a checkable action with a label in to the menu

        Args:
            label (str): Label for the checkbox

        Returns:
            QAction: The generated QAction
        """

        action = QAction(label, parent=self)
        action.setCheckable(True)
        action.setChecked(True)

        if self._sorted:
            insert_index = bisect.bisect_left(
                [action.text() for action in self._filter_actions], label
            )
            try:
                before = self._filter_actions[insert_index]
            except IndexError:
                before = None
        else:
            before = None
            insert_index = len(self._filter_actions)

        self.insertAction(before, action)
        self._filter_actions.insert(insert_index, action)

        return action

    def remove_filter_action(self, action: QAction) -> None:
        """Removes the action from the menu

        Args:
            action (QAction): Action to be removed
        """

        self.removeAction(action)
        self._filter_actions.remove(action)
        action.deleteLater()


class AbstractQualityErrorFilter(QObject):
    """Abstract Base Class for Quality Error Filters

    Attributes:
        menu (QMenu): Bounded QMenu to use as a submenu in QualityErrorsTreeFilterMenu
    """

    filters_changed = pyqtSignal()

    def __init__(self, title: str, parent: Optional[QObject] = None) -> None:
        """Inits a QualityErrorFilter

        Holds a sub menu in the actual QualityErrorsTreeFilterMenu opened by the filter
        button.

        Inherited classes must implement accept_row method that is called by
        filter proxy model.

        self._accepted_values holds current values that are checked by bounded checkable
        actions.

        Args:
            title (str): Label text for the QMenu
            parent (Optional[QObject], optional): QtParent object. Defaults to None.
        """

        super().__init__(parent=parent)
        self._accepted_values: Set[Any] = set()
        self._filter_value_action_map: Dict[Hashable, QAction] = {}

        self.menu = FilterMenu(title)

    @abstractmethod
    def accept_row(self, item_type: QualityErrorTreeItemType, item_value: Any) -> bool:
        """The actual filter function that inherited classes should be implemented.

        Args:
            item_type (QualityErrorTreeItemType): Type of the item_value
            item_value (Any): Item to be tested by the filter

        Raises:
            NotImplementedError: Should be implemented by inherited classes

        Returns:
            bool: Is the item accepted by the filter
        """
        raise NotImplementedError()

    def update_filter_from_errors(
        self, quality_errors: List["QualityErrorsByPriority"]
    ) -> None:
        """Updates filters dynamically from a given list of quality errors.

        Should be implemented in inherited classes if feature is wanted.

        Args:
            quality_errors (List[&quot;QualityErrorsByPriority&quot;]): List of errors
              to use to update filters

        Raises:
            NotImplementedError: Should be implemented in inherited classes
        """
        raise NotImplementedError()

    def _sync_filtered(self, value: Any, checked: bool) -> None:
        """Syncs accepted filter values

        Should be connected to checkable action's toggle signal with functool.partial
        where value will be fixed.

        Args:
            value (Any): Value to be added or removed from the accepted value set.
            checked (bool): Is action is checked or not.
        """

        if checked:
            self._accepted_values.add(value)
        else:
            self._accepted_values.remove(value)

        self.filters_changed.emit()

    def _refresh_filters(self, new_filters: Dict[Any, str]) -> None:
        """Adds filters not yet present and removes filters not present anymore.

        Args:
            new_filters (Dict[Any, str]): Filters that should be available after
              the refresh. Dict keys are the filter values and dict values are labels
              for the filters.
        """

        new_values = set(new_filters.keys())
        current_values = set(self._filter_value_action_map.keys())

        values_to_be_added = new_values - current_values
        values_to_be_removed = current_values - new_values

        for filter_value in values_to_be_removed:
            self._remove_filter_item(filter_value)

        for filter_value in values_to_be_added:
            self._add_filter_item(filter_value, new_filters[filter_value])

    def _add_filter_item(self, filter_value: Any, filter_label: str) -> None:
        """Adds a filter item to the filter

        Args:
            filter_value (Any): Value to be used when filtering
            filter_label (str): Label text shown in the menu
        """

        self._accepted_values.add(filter_value)

        action = self.menu.add_checkable_action(filter_label)
        self._filter_value_action_map[filter_value] = action
        action.toggled.connect(partial(self._sync_filtered, filter_value))

        self.filters_changed.emit()

    def _remove_filter_item(self, filter_value: Any) -> None:
        """Removes the filter item

        Args:
            filter_value (Any): The Filter Value that should be removed from the filter
        """

        action = self._filter_value_action_map.pop(filter_value)
        action.deleteLater()
        self.menu.remove_filter_action(action)

        if filter_value in self._accepted_values:
            self._accepted_values.remove(filter_value)

        self.filters_changed.emit()


class ErrorTypeFilter(AbstractQualityErrorFilter):
    """Filter for the Error Type of the Quality errors.

    This is a static filter that shows always all the defined error types.
    """

    _accepted_values: Set[QualityErrorTreeItemType]

    def __init__(self) -> None:
        super().__init__(ERROR_TYPE_FILTER_MENU_LABEL)
        self.menu.set_select_all_section_enabled(True)
        self.menu.set_sorted(True)

        self._refresh_filters(ERROR_TYPE_LABEL)

    def accept_row(self, item_type: QualityErrorTreeItemType, item_value: Any) -> bool:
        if item_type == QualityErrorTreeItemType.ERROR:
            return cast(QualityError, item_value).error_type in self._accepted_values

        return True


class FeatureTypeFilter(AbstractQualityErrorFilter):
    """Filter for the Feature Type of the Quality errors.

    Adds filterable feature types dynamically based on the current errors received.
    update_filter_from_errors slot should be connected to the signal that transmits
    received errors.
    """

    _accepted_values: Set[str]

    def __init__(self) -> None:
        super().__init__(FEATURE_TYPE_FILTER_MENU_LABEL)
        self.menu.set_select_all_section_enabled(True)
        self.menu.set_sorted(True)

    def accept_row(self, item_type: QualityErrorTreeItemType, item_value: Any) -> bool:
        if item_type == QualityErrorTreeItemType.FEATURE_TYPE:
            return cast(str, item_value) in self._accepted_values

        return True

    def update_filter_from_errors(
        self, quality_errors: List["QualityErrorsByPriority"]
    ) -> None:
        """

        Args:
            quality_errors (List[&quot;QualityErrorsByPriority&quot;]): _description_
        """
        feature_types_in_errors = {  # Dict[filter_value, filter_label]
            errors.feature_type: errors.feature_type
            for errors_by_feature_type in quality_errors
            for errors in errors_by_feature_type.errors
        }

        self._refresh_filters(feature_types_in_errors)


class AttributeFilter(AbstractQualityErrorFilter):
    """Filter for the Attribute Errors of the Quality errors.

    Adds filterable Attribute names dynamically based on the current errors received.
    update_filter_from_errors slot should be connected to the signal that transmits
    received errors.
    """

    _accepted_values: Set[str]

    def __init__(self) -> None:
        super().__init__(ATTRIBUTE_NAME_FILTER_MENU_LABEL)
        self.menu.set_select_all_section_enabled(True)
        self.menu.set_sorted(True)

    def accept_row(self, item_type: QualityErrorTreeItemType, item_value: Any) -> bool:
        if item_type == QualityErrorTreeItemType.ERROR:
            attribute_name = cast(QualityError, item_value).attribute_name
            if attribute_name:
                return attribute_name in self._accepted_values

        return True

    def update_filter_from_errors(
        self, quality_errors: List["QualityErrorsByPriority"]
    ) -> None:
        attribute_names_in_errors = {  # Dict[filter_value, filter_label]
            error.attribute_name: error.attribute_name
            for errors_by_priority in quality_errors
            for errors_by_feature_type in errors_by_priority.errors
            for errors_by_feature in errors_by_feature_type.errors
            for error in errors_by_feature.errors
            if error.attribute_name
        }

        self._refresh_filters(attribute_names_in_errors)



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import Iterable, List, Optional

from qgis.core import QgsAnnotationLayer, QgsCoordinateReferenceSystem, QgsProject

from quality_result_gui.api.types.quality_error import QualityError
from quality_result_gui.configuration import QualityLayerStyleConfig
from quality_result_gui.quality_layer import QualityErrorLayer
from quality_result_gui.ui.quality_error_tree_view import SelectionType
from quality_result_gui.utils import layer_utils

LOGGER = logging.getLogger(__name__)


class QualityErrorVisualizer:
    """
    Class for visualizing quality errors on map canvas.
    """

    ID_PREFIX_FOR_SELECTED = "selected-"
    style_config: Optional["QualityLayerStyleConfig"] = None

    def __init__(
        self,
        crs: QgsCoordinateReferenceSystem,
        style_config: Optional["QualityLayerStyleConfig"] = None,
    ) -> None:
        self._crs = crs
        self._selected_quality_error: Optional[QualityError] = None
        self.style_config = style_config

        self._quality_error_layer = QualityErrorLayer()

    def toggle_visibility(self, show_errors: bool) -> None:
        if show_errors is True:
            self.show_errors()
        else:
            self.hide_errors()

    def add_new_errors(self, quality_errors: Iterable[QualityError]) -> None:
        for quality_error in quality_errors:
            self._quality_error_layer.add_or_replace_annotation(
                quality_error, use_highlighted_style=False
            )

    def remove_errors(self, quality_errors: Iterable[QualityError]) -> None:
        self._quality_error_layer.remove_annotations(quality_errors)

    def on_error_selected(
        self, quality_error: QualityError, selection_type: SelectionType
    ) -> None:
        preserve_scale = selection_type == SelectionType.RightClick

        self.zoom_to_geometries_and_flash(
            [quality_error], preserve_scale=preserve_scale
        )

        self.refresh_selected_error(quality_error)

    def refresh_selected_error(
        self,
        selected_quality_error: QualityError,
    ) -> None:
        self._remove_selected_error()
        self._selected_quality_error = selected_quality_error

        self._quality_error_layer.add_or_replace_annotation(
            selected_quality_error,
            use_highlighted_style=True,
            id_prefix=self.ID_PREFIX_FOR_SELECTED,
        )

    def show_errors(self) -> None:
        layer = self._get_or_create_layer()
        self.override_quality_layer_style()
        layer_utils.set_visibility_checked(layer, True)

    def hide_errors(self) -> None:
        layer = self._get_or_create_layer()
        layer_utils.set_visibility_checked(layer, False)

    def initialize_quality_error_layer(self, visible: bool = True) -> None:
        self.remove_quality_error_layer()
        layer = self._get_or_create_layer()
        layer_utils.set_visibility_checked(layer, visible)

    def remove_quality_error_layer(self) -> None:
        layer = self._quality_error_layer.find_layer_from_project()

        if layer is not None:
            QgsProject.instance().removeMapLayer(layer.id())

    def zoom_to_geometries_and_flash(
        self, quality_errors: List[QualityError], preserve_scale: bool = False
    ) -> None:
        if len(quality_errors) > 0:
            layer_utils.zoom_to_geometries_and_flash(
                [error.geometry for error in quality_errors],
                self._crs,
                preserve_scale,
                min_extent_height=20,
            )

    def override_quality_layer_style(self) -> None:
        if self.style_config:
            self._quality_error_layer.override_style(self.style_config)

    def _get_or_create_layer(self) -> QgsAnnotationLayer:
        layer = self._quality_error_layer.find_layer_from_project()

        if layer is None:
            layer = self._quality_error_layer.get_annotation_layer()
            QgsProject.instance().addMapLayer(layer, False)
            QgsProject.instance().layerTreeRoot().insertLayer(0, layer)

        return layer

    def _remove_selected_error(self) -> None:
        if self._selected_quality_error is not None:
            self._quality_error_layer.remove_annotations(
                [self._selected_quality_error], id_prefix=self.ID_PREFIX_FOR_SELECTED
            )
            self._selected_quality_error = None
